/*****************************************************************************
 *
 * FILE:	cockatrice.c
 * DESCRIPTION:	simple CGI library
 * DATE:	Sat, Mar 16 2002
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2002 졿Kouichi ABE (WALL), All rights reserved.
 * $Id: cockatrice.c,v 1.17 2002/07/22 08:36:13 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_DIRENT_H
#include <dirent.h>
#else
#define dirent	direct
#if	HAVE_SYS_NDIR_H
#include <sys/ndir.h>
#endif	/* HAVE_SYS_NDIR_H */
#if	HAVE_SYS_DIR_H
#include <sys/dir.h>
#endif	/* HAVE_SYS_DIR_H */
#if	HAVE_NDIR_H
#include <ndir.h>
#endif	/* HAVE_NDIR_H */
#endif	/* HAVE_DIRENT_H */
#include <sys/types.h>
#include <sys/mman.h>
#include <signal.h>
#include <errno.h>
#if	USE_WKF
#include <wkf.h>
#endif	/* USE_WKF */
#include "cockatrice.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of global variables
 *
 *****************************************************************************/
cgierr_t	cgi_errno;
#if	DEBUG
FILE *		debugLog;
#endif	/* DEBUG */


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#define	LIBRARY_NAME	"COCKATRICE"

#define	MULTIPART_FORM	"multipart/form-data"
#define	MULTIPART_MIXED	"multipart/mixed"

#define	THRESHOLD_SIZE	(1024*2)

typedef struct _query {	/* query structure */
  String	string;
  size_t	length;
  int		fd;	/* used by mmap() */
} Query;


/*****************************************************************************
 *
 *	definition of local functions
 *
 *****************************************************************************/
static void	cgiDone(void);
static void	redirectURI(const String);
static int	removeTempDir(const String);
static int	readData(void);
static int	parseQuery(void);
static void	trap(int);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static CGI *	cgi;	/* cgi object */
static Query	query;	/* received data */


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * FUNCTION NAME:	newCGI
 * DESCRIPTION:		create new CGI object
 * ARGUMENTS:		DTD for HTML authoring
 * RETURN VALUE:	CGI object
 */
CGI *
newCGI(dtd)
	dtd_t	dtd;
{
  static bool_t		once = FALSE;
#if	HAVE_MKDTEMP
  static char		tempdir[] = "/tmp/cockatrice.XXXXXX";
#else
  int			n;
#endif	/* HAVE_MKDTEMP */
  String		p;
  struct sigaction	act;

  if (once) {	/* called only once */
    cgi_errno = CGIERR_ALREDY_MADE_OBJECT;
    return NULL;
  }
  else {
    once = TRUE;
  }

  cgi = (CGI *)calloc(1, sizeof(CGI));
  if (cgi == NULL) {
    cgi_errno = CGIERR_NO_ENOUGH_MEMORY;
    return NULL;
  }
#if	DEBUG
  debugLog = fopen("/tmp/cockatrice-debug.log", "w");
#endif	/* DEBUG */

  /* for QUERY_STRING */
  paramInit();

  /* for Cookie */
  cookieInit();

  /* for Environment Variables */
  if (envInit() == -1) {
    return NULL;
  }

  /* for Date */
  dateInit();

  /* temporary directory */
#if	HAVE_MKDTEMP
  cgi->tmpdir = mkdtemp(tempdir);
  if (cgi->tmpdir == NULL) {
    cgi_errno = CGIERR_NO_TEMPORARY_DIR;
    return NULL;
  }
#else
  n = asprintf(&cgi->tmpdir, "/tmp/cockatrice.%06d", getpid());
  if (n == -1) {
    cgi_errno = CGIERR_NO_TEMPORARY_DIR;
    return NULL;
  }
  (void)mkdir(cgi->tmpdir, 0700);
#endif	/* HAVE_MKDTEMP */

  /**/
  query.string	= NULL;
  query.length	= 0;
  query.fd	= -1;

  p = getenv("SCRIPT_NAME");
  if (p != NULL) {
    cgi->script_name = strdup(p);
  }
  cgi->param			= (void *)(paramGet);
  cgi->params			= (void *)(paramsGet);
  /**/
  cgi->cookie			= (void *)(cookieGet);
  cgi->setCookie.Name		= (void *)(cookieSetName);
  cgi->setCookie.Comment	= (void *)(cookieSetComment);
  cgi->setCookie.Domain		= (void *)(cookieSetDomain);
  cgi->setCookie.MaxAge		= (void *)(cookieSetMaxAge);
  cgi->setCookie.Path		= (void *)(cookieSetPath);
  cgi->setCookie.Secure		= (void *)(cookieSetSecure);
  /**/
  cgi->session.openDB		= (void *)(sessionOpenDB);
  cgi->session.closeDB		= (void *)(sessionCloseDB);
  cgi->session.setExpires	= (void *)(sessionSetExpires);
  cgi->session.getExpires	= (void *)(sessionGetExpires);
  cgi->session.generateID	= (void *)(sessionGenerateID0);
  cgi->session.generateID1	= (void *)(sessionGenerateID1);
  cgi->session.generateID2	= (void *)(sessionGenerateID2);
  cgi->session.validateID	= (void *)(sessionValidateID);
  cgi->session.deleteID		= (void *)(sessionDeleteID);
  cgi->session.countValidID	= (void *)(sessionCountValidID);
  /**/
  cgi->upload.open		= (void *)(openUploadFile);
  cgi->upload.store		= (void *)(storeUploadFile);
  cgi->upload.digest		= (void *)(digestUploadFile);
  /**/
  cgi->getenv			= (void *)(getenv);
  cgi->env.server_software	= (void *)(envServerSoftware);
  cgi->env.server_name		= (void *)(envServerName);
  cgi->env.gateway_interface	= (void *)(envGatewayInterface);
  cgi->env.server_protocol	= (void *)(envServerProtocol);
  cgi->env.server_port		= (void *)(envServerPort);
  cgi->env.request_method	= (void *)(envRequestMethod);
  cgi->env.script_name		= (void *)(envScriptName);
  cgi->env.path_info		= (void *)(envPathInfo);
  cgi->env.path_translated	= (void *)(envPathTranslated);
  cgi->env.remote_host		= (void *)(envRemoteHost);
  cgi->env.remote_addr		= (void *)(envRemoteAddr);
  cgi->env.remote_port		= (void *)(envRemotePort);
  cgi->env.auth_type		= (void *)(envAuthType);
  cgi->env.remote_user		= (void *)(envRemoteUser);
  cgi->env.remote_ident		= (void *)(envRemoteIdent);
  cgi->env.accept		= (void *)(envHttpAccept);
  cgi->env.accept_encoding	= (void *)(envHttpAcceptEncoding);
  cgi->env.accept_language	= (void *)(envHttpAcceptLanguage);
  cgi->env.user_agent		= (void *)(envHttpUserAgent);
  cgi->env.referer		= (void *)(envHttpReferer);
  cgi->env.document_root	= (void *)(envDocumentRoot);
  cgi->env.server_root		= (void *)(envServerRoot);
  cgi->env.server_admin		= (void *)(envServerAdmin);
  /**/
  cgi->util.lock		= (void *)(lock);
  cgi->util.unlock		= (void *)(unlock);
  cgi->util.escape		= (void *)(escapeURI);
  cgi->util.unescape		= (void *)(unescapeURI);
  cgi->util.digest		= (void *)(digestString);
  /**/
  cgi->date.today		= (void *)(getToday);
  cgi->date.year		= (void *)(getYearOfToday);
  cgi->date.month		= (void *)(getMonthOfToday);
  cgi->date.day			= (void *)(getDayOfToday);
  cgi->date.wday		= (void *)(getWDayOfToday);
  cgi->date.isFestival		= (void *)(isFestival);
  cgi->date.getWDay		= (void *)(getDayOfTheWeek);
  cgi->date.time		= (void *)(getTime);
  /**/
  cgi->check.empty		= (void *)(checkEmpty);
  cgi->check.digit		= (void *)(checkDigit);
  cgi->check.alphabet		= (void *)(checkAlphabet);
  cgi->check.username		= (void *)(checkUserName);
  cgi->check.hostname		= (void *)(checkHostName);
  cgi->check.domainname		= (void *)(checkDomainName);
  cgi->check.email		= (void *)(checkEMail);
  cgi->check.range		= (void *)(checkRange);
  cgi->check.ipaddr		= (void *)(checkIPAddr);
  cgi->check.ipaddr2		= (void *)(checkNetAddr);
  cgi->check.ipaddr3		= (void *)(checkIPandPort);
  /**/
  cgi->done			= (void *)(cgiDone);
  /**/
  cgi->header			= (void *)(sendResponseHeader);
  cgi->redirect			= (void *)(redirectURI);
  /**/
  cgi->html			= newHTML401(dtd);
  if (cgi->html == NULL) {
    cgi_errno = CGIERR_NO_ENOUGH_MEMORY;
    return NULL;
  }

  /* set signal trap */
  sigemptyset(&act.sa_mask);
  act.sa_flags	 = 0;
  act.sa_handler = (void *)trap;
  sigaction(SIGINT,  &act, NULL);
  sigaction(SIGQUIT, &act, NULL);
  sigaction(SIGTERM, &act, NULL);
  sigaction(SIGBUS,  &act, NULL);
  sigaction(SIGSEGV, &act, NULL);
  sigaction(SIGPIPE, &act, NULL);

  if (readData() == -1) {
    return NULL;
  }

  cgi_errno = CGIERR_NO_ERROR;
  return cgi;
}

/*
 * FUNCTION NAME:	cgiDone
 * DESCRIPTION:		cleanup all resources
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
static void
cgiDone(void)
{
  sigset_t	set, oset;

  html401Done();

  /* block all signals */
  sigfillset(&set);
  sigprocmask(SIG_BLOCK, &set, &oset);

  if (query.string != NULL) {
    if (query.fd > 0) {
      munmap((String)query.string, (size_t)query.length);
      close(query.fd);
    }
    else {
      free((String)query.string);
    }
  }

  unlockAll();
  cookieRelease();
  paramRelease();
  envRelease();
  sessionCloseDB();
  closeUploadFiles();

  if (cgi->script_name != NULL) {
    free((String)cgi->script_name);
  }

#if	DEBUG
  if (removeTempDir(cgi->tmpdir) == -1) {
    fprintf(debugLog, "errno = %d\n", errno);
  }
#else
  removeTempDir(cgi->tmpdir);
#endif	/* DEBUG */
#ifndef	HAVE_MKDTEMP
  free((String)cgi->tmpdir);
#endif	/* HAVE_MKDTEMP */
  free((CGI *)cgi);

  /* restore signal mask */
  sigprocmask(SIG_SETMASK, &oset, NULL);
#if	DEBUG
  if (debugLog != NULL) {
    fclose(debugLog);
  }
#endif	/* DEBUG */
}

/*
 * FUNCTION NAME:	redirectURI
 * DESCRIPTION:		flush out to Location response-header
 * ARGUMENTS:		URI
 * RETURN VALUE:	none
 */
static void
redirectURI(uri)
	const String	uri;
{
  fputs("Status: 302 Moved" CRLF, stdout);
  fprintf(stdout, "Location: %s" CRLF, uri);
  fputs(CRLF, stdout);
}


/*****************************************************************************
 *****************************************************************************/

/*
 * FUNCTION NAME:	removeTempDir
 * DESCRIPTION:		remove a temporary directory
 * ARGUMENTS:		directory
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
removeTempDir(dirname)
	const String	dirname;
{
  DIR *			dfd;
  struct dirent *	dp;
  String		fullpath;

  dfd = opendir(dirname);
  if (dfd == NULL) {
    return -1;
  }

  while ((dp = readdir(dfd)) != NULL) {
    int	n = asprintf(&fullpath, "%s/%s", dirname, dp->d_name);
    if (n == -1) {
      continue;
    }
    if ((dp->d_name[0] == '.' && strlen(dp->d_name) == 1) ||
	(dp->d_name[0] == '.' && dp->d_name[1] == '.' &&
	 strlen(dp->d_name) == 2)) {
      continue;
    }
    else if (dp->d_type == DT_DIR) {
      removeTempDir(fullpath);
    }
    else {
      unlink(fullpath);
    }
    free((String)fullpath);
  }
  closedir(dfd);

  return rmdir(dirname);
}


/*****************************************************************************
 *****************************************************************************/

/*
 * FUNCTION NAME:	readData
 * DESCRIPTION:		read CGI data and set it to qeury buffer
 * ARGUMENTS:		none
 * RETURN VALUE:	Successful: 0, othereise: -1
 */
static int
readData(void)
{
  String	method	  = NULL;
  bool_t	multipart = FALSE;

  /* get REQUEST_METHOD */
  method = envRequestMethod();
  if (method == NULL) {
    cgi_errno = CGIERR_NO_REQUEST_METHOD;
    return -1;
  }

  /* GET */
  if (strcasecmp(method, "GET") == 0) {
    String	query_string = getenv("QUERY_STRING");

    if (query_string == NULL) {
      cgi_errno = CGIERR_NO_QUERY_STRING;
      return -1;
    }

    query.string = strdup(query_string);
    if (query.string == NULL) {
      cgi_errno = CGIERR_NO_ENOUGH_MEMORY;
      return -1;
    }
    query.length = strlen(query_string);
  }
  /* POST */
  else if (strcasecmp(method, "POST") == 0) {
    String	content_length	= getenv("CONTENT_LENGTH");
    String	content_type	= getenv("CONTENT_TYPE");

    if (content_length == NULL) {
      cgi_errno = CGIERR_NO_CONTENT_LENGTH;
      return -1;
    }
    query.length = (size_t)atoi(content_length);

    if (content_type != NULL &&
	strstr(content_type, MULTIPART_FORM) != NULL) {
      multipart = TRUE;
    }
    else if (query.length < THRESHOLD_SIZE) {
      query.string = (String)calloc(query.length + 1, sizeof(char));
      if (query.string == NULL) {
	cgi_errno = CGIERR_NO_ENOUGH_MEMORY;
	return -1;
      }
      if (fread(query.string, 1, query.length, stdin) != query.length) {
	cgi_errno = CGIERR_INTERNAL_LIBRARY_ERROR;
	return -1;
      }
    }
    else {
      char	tempfile[] = "/tmp/cockatrice.XXXXXX";
      char	buf[BUFSIZ];
      int	len = 0;
      int	n;

      query.fd = mkstemp(tempfile);
      if (query.fd == -1) {
	cgi_errno = CGIERR_TMPFILE_OPEN_FAILURE;
	return -1;
      }
      unlink(tempfile);	/* the file is removed when it's closed. */

      memset(buf, 0, sizeof(buf));
      while ((n = read(0, buf, sizeof(buf))) > 0 && len < query.length) {
	write(query.fd, buf, (size_t)n);
	memset(buf, 0, sizeof(buf));
	len += n;
      }
      if (len != query.length) {
	cgi_errno = CGIERR_INTERNAL_LIBRARY_ERROR;
	return -1;
      }

      query.string = (String)mmap(0,			/* address */
				 (size_t)query.length,	/* size */
				 PROT_READ|PROT_WRITE,	/* protection */
				 MAP_PRIVATE,	/* type of the mapped object */
				 query.fd,		/* file descriptor */
				 (off_t)0);		/* offset */

      if (query.string == MAP_FAILED) {
	close(query.fd);
	query.fd = -1;
	cgi_errno = CGIERR_INTERNAL_LIBRARY_ERROR;
	return -1;
      }
    }
  }
  /* error */
  else {
    cgi_errno = CGIERR_ILLEGAL_METHOD_TYPE;
    return -1;
  }

  /* cookie */
  cookieParse();

  /* parse */
  if (multipart) {
    return parseMultiPartForm();
  }
  else {
    return parseQuery();
  }

  return 0;	/* not reached */
}

/*
 * FUNCTION NAME:	parseQuery
 * DESCRIPTION:		parse CGI query data
 * ARGUMENTS:		none
 * RETURN VALUE:	Successfule: 0, otherwise: -1
 */
static int
parseQuery(void)
{
  String	str	= query.string;
  String	name	= NULL;
  String	value	= NULL;

#if	USE_WKF
  wkfSetLineEndCode(LE_LF);
#endif	/* USE_WKF */

  while ((str = strtok(str, "&")) != NULL) {
#if	USE_WKF
    String		Kname;
    String		Kvalue;
    unsigned int	len;
#endif	/* USE_WKF */
    String		p;

    name = unescapeURI(str);	/* original string is broken */
    p	 = (String)strchr(str, '=');
    if (p != NULL) {
      *p++  = EOL;
      value = unescapeURI(chomp(p));
#if	USE_WKF
      len	= strlen(name) + 1;
      Kname	= (String)calloc(len, sizeof(char));
      if (Kname != NULL &&
	  wkfConvertKanjiCodeOfString(KC_UNKNOWN, name,
				      KC_EUC, Kname, len) != CONV_ERR) {
	name = Kname;
      }
      len	= strlen(value) + 1;
      Kvalue	= (String)calloc(len, sizeof(char));
      if (Kvalue != NULL &&
	  wkfConvertKanjiCodeOfString(KC_UNKNOWN, value,
				      KC_EUC, Kvalue, len) != CONV_ERR) {
	value = Kvalue;
      }
#endif	/* USE_WKF */
      paramPut(name, value);
#if	USE_WKF
      if (Kname != NULL) {
	free((String)Kname);
      }
      if (Kvalue != NULL) {
	free((String)Kvalue);
      }
#endif	/* USE_WKF */
    }
    str = NULL;	/* DON'T REMOVE THIS LINE. */
  }

  return 0;
}


/*****************************************************************************
 *****************************************************************************/

/*
 * Function Name:	trap
 * Description:		trap all signals
 * Arguments:		signal
 * Return Value:	never return
 */
static void
trap(sig)
	int	sig;
{
  fprintf(stderr, "Cockatrice caught a signal(%d)\n", sig);
  cgiDone();
  exit(1);
}

