/*****************************************************************************
 *
 * FILE:	upload.c
 * DESCRIPTION:	upload manipulator for simple CGI library
 * DATE:	Tue, Mar 19 2002
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.com/cover.html
 * COPYRIGHT:	(C) 2002 졿Kouichi ABE (WALL), All rights reserved.
 * $Id: upload.c,v 1.7 2002/06/19 11:47:23 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <sys/types.h>
#include <md5.h>
#include <sha.h>
#include "cockatrice.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#define	CONTENT_DISPOSITION		"Content-Disposition"
#define	CONTENT_TYPE			"Content-Type"
#define	CONTENT_TRANSFER_ENCODING	"Content-Transfer-Encoding"

typedef struct _content_header {	/* RFC1867 */
  struct {	/* Content-Disposition ... RFC1806 */
    String	type;	/* inline/attachment/form-data */
    String	name;
    String	filename;
  } Disposition;
  struct {	/* Content-Type: type/subtype; paramter */
    String	media_type;
    String	parameter;
  } Type;
  /* 7bit/8bit/binary/quoted-printable/base64/ietf-token/x-token */
  String	TransferEncoding;
} Header;

typedef struct _ListManager {
  String		name;
  FILE *		fp;
  struct _ListManager *	next;
} ListManager;

#define	XFree(x)	if((x)!=NULL){free((x));(x)=NULL;}


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static String	getSeparator(const String);
static int	readHeader(Header *);
static String	getData(const String);
static String	getValue(String);
static int	readEntity(const String, const String);
static String	getEntity(const String);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static ListManager *	files = NULL;	/* upload file list */


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
Content-type: multipart/form-data; boundary=------------------------------29531351678801353387200
*/

/*
 * FUNCTION NAME:	parseMultiPartForm
 * DESCRIPTION:		parse multipart form data
 * ARGUMENTS:		none
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
int
parseMultiPartForm(void)
{
  String	separator = getSeparator(getenv("CONTENT_TYPE"));
  Header	content;

  if (separator == NULL) {
    cgi_errno = CGIERR_ILLEGAL_MULTIPART_FORM;
    return -1;
  }

  content.Disposition.type	= NULL;
  content.Disposition.name	= NULL;
  content.Disposition.filename	= NULL;
  content.Type.media_type	= NULL;
  content.Type.parameter	= NULL;
  content.TransferEncoding	= NULL;

  while (!feof(stdin)) {
    if (readHeader(&content) == -1) {
      return -1;
    }

    if (content.Disposition.name != NULL) {
      if (content.Disposition.filename != NULL) {
	paramPut(content.Disposition.name, content.Disposition.filename);
	readEntity(separator, content.Disposition.name);
      }
      else {
	String	value = getEntity(separator);

	paramPut(content.Disposition.name, value);
      }
    }
    XFree(content.Disposition.type);
    XFree(content.Disposition.name);
    XFree(content.Disposition.filename);
    XFree(content.Type.media_type);
    XFree(content.Type.parameter);
    XFree(content.TransferEncoding);
  }

  return 0;
}

/*
 * FUNCTION NAME:	openUploadFile
 * DESCRIPTION:		open upload file
 * ARGUMENTS:		upload filename
 * RETURN VALUE:	Successful: file pointer, otherwise: NULL
 */
FILE *
openUploadFile(filename)
	const String	filename;
{
  register ListManager *	lmp;

  for (lmp = files; lmp != NULL; lmp = lmp->next) {
    if (strcmp(filename, lmp->name) == 0) {
      return lmp->fp;
    }
  }

  return NULL;
}

/*
 * FUNCTION NAME:	closeUploadFiles
 * DESCRIPTION:		close all upload files
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
void
closeUploadFiles(void)
{
  register ListManager *	lmp;
  register ListManager *	tmp;

  for (lmp = files; lmp != NULL; lmp = tmp) {
    tmp = lmp->next;
    XFree(lmp->name);
    fclose(lmp->fp);
    free((ListManager *)lmp);
  }
}


/*****************************************************************************
 *****************************************************************************/

/*
 * FUNCTION NAME:	getSeparator
 * DESCRIPTION:		get a separator of multipart form data
 * ARGUMENTS:		Content-Type line
 * RETURN VALUE:	Successful: separator, otherwise: NULL
 */
static String
getSeparator(line)
	const String	line;
{
  String	p;
  String	separator;
  int		n = -1;

  p = (String)strstr(line, "boundary=");
  if (p != NULL) {
    n = asprintf(&separator, "--%s", p + 9);
  }

  return n == -1 ? NULL : separator;
}

/*
 * FUNCTION NAME:	readHeader
 * DESCRIPTION:		read a header of multipart form data
 * ARGUMENTS:		header received buffer
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
readHeader(content)
	Header *	content;
{
  char	buf[BUFSIZ];

  memset(buf, 0, sizeof(buf));
  while (!feof(stdin)) {
    if (fgets(buf, sizeof(buf), stdin) == NULL) {
      return 0;
    }
    if (buf[0] == CR && buf[1] == LF) {
      return 0;
    }
    else if (strncasecmp(buf, CONTENT_DISPOSITION, 19) == 0) {
      String	args[3];
      String *	ap;
      String	disposition = getData(buf);
      int	i;

      if (disposition == NULL) {
	return -1;
      }

      for (i = 0; i < 3; i++) {
	args[i] = NULL;
      }
      for (i = 0, ap = args; (*ap = strsep(&disposition, ";")) != NULL; i++) {
	args[i] = *ap;
	if (**ap != EOL) {
	  if (++ap >= &args[3]) {
	    break;	/* error: arguments overflow */
	  }
	}
      }
      if (args[0] != NULL) {
	content->Disposition.type = getValue(args[0]);
      }
      if (args[1] != NULL) {
	content->Disposition.name = getValue(args[1]);
      }
      if (args[2] != NULL) {
	content->Disposition.filename = getValue(args[2]);
      }
      free((String)disposition);
    }
    else if (strncasecmp(buf, CONTENT_TYPE, 12) == 0) {
      String	type = getData(buf);
      String	p;

      if (type == NULL) {
	return -1;
      }

      p = strchr(type, ';');
      if (p != NULL) {
	*p = EOL;
	for (++p; *p == SPC; p++) {	/* remove white space */
	  ;	/* nothing */
	}
	content->Type.parameter = strdup(p);
      }
      content->Type.media_type = strdup(type);
      free((String)type);
    }
    else if (strncasecmp(buf, CONTENT_TRANSFER_ENCODING, 25) == 0) {
      content->TransferEncoding = getData(buf);
    }
    else {
      /* ignored other header fields */
    }
    memset(buf, 0, sizeof(buf));
  }

  return -1;
}

/*
 * FUNCTION NAME:	getData
 * DESCRIPTION:		get data of content header field
 * ARGUMENTS:		content header field
 * RETURN VALUE:	Successful: data, otherwise: NULL
 */
static String
getData(field)
	String	field;
{
  String	p;

  p = strchr(field, ':');
  if (p != NULL) {
    for (++p; *p == SPC; p++) {	/* remove white space */
      ;	/* nothing to do */
    }
    return strdup(p);
  }

  return NULL;
}

/*
 * FUNCTION NAME:	getValue
 * DESCRIPTION:		get a value of data field
 * ARGUMENTS:		data field
 * RETURN VALUE:	Successful: value, otherwise: NULL
 */
static String
getValue(field)
	String	field;
{
  String	value;
  String	p;

  p = strchr(field, '=');
  if (p != NULL) {	/* name=value form */
    value = ++p;
    p	  = strchr(value, '"');
    if (p != NULL) {	/* double quotation */
      value = ++p;
      p = strrchr(value, '"');
      if (p != NULL) {
	*p = EOL;
	return *value == EOL ? NULL : strdup(value);
      }
    }
    return NULL;
  }

  while (*field == SPC) {
    field++;
  }

  return *field == EOL ? NULL : field;
}

/*
 * FUNCTION NAME:	readEntity
 * DESCRIPTION:		read entity of multipart form data
 * ARGUMENTS:		separator, name
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
readEntity(separator, name)
	const String	separator;
	const String	name;
{
  size_t	slen = strlen(separator);
  FILE *	fout;
  ListManager *	lmp;
  char		buf[BUFSIZ];
  char		c;
  int		i;
  size_t	size = 0;

  fout = tmpfile();
  if (fout == NULL) {
    cgi_errno = CGIERR_TMPFILE_OPEN_FAILURE;
    return -1;
  }

  lmp = (ListManager *)calloc(1, sizeof(ListManager));
  if (lmp == NULL) {
    fclose(fout);
    cgi_errno = CGIERR_NO_ENOUGH_MEMORY;
    return -1;
  }
  lmp->name = strdup(name);
  lmp->fp   = fout;
  lmp->next = files;
  files	    = lmp;

  i = 0;
  while (!feof(stdin)) {
    if (fread(&c, sizeof(char), 1, stdin) > 0) {
      buf[i] = c;
      i++;
      if (i >= BUFSIZ) {
	size += fwrite(buf, sizeof(char), BUFSIZ, fout);
	i = 0;
      }
      else if (c == LF) {
	buf[i]	= EOL;
	if (strncmp(buf, separator, slen) == 0) {
	  (void)fseek(fout, 0L, SEEK_SET);
	  /* remove CRLF in last line */
	  return ftruncate(fileno(fout), (off_t)(size - 2));
	}
	size += fwrite(buf, sizeof(char), (size_t)i, fout);
	i = 0;
      }
    }
  }

  return -1;
}

/*
 * FUNCTION NAME:	getEntity
 * DESCRIPTION:		get entity of multipart form data
 * ARGUMENTS:		separator
 * RETURN VALUE:	Successful: entity, otherwise: NULL
 */
static String
getEntity(separator)
	const String	separator;
{
  char		buf[BUFSIZ];
  size_t	slen	= strlen(separator);
  size_t	size	= 0;
  String	value	= calloc(1, sizeof(char));

  if (value == NULL) {
    return NULL;
  }
  value[0] = EOL;

  memset(buf, 0, sizeof(buf));
  while (fgets(buf, sizeof(buf), stdin) != NULL) {
    size_t	len;

    if (strncmp(buf, separator, slen) == 0) {
      chomp(value);
      return *value == EOL ? NULL : value;
    }
    len	  = strlen(buf);
    size += len;
    value = realloc(value, size + 1);
    if (value == NULL) {
      free((String)value);
      cgi_errno = CGIERR_NO_ENOUGH_MEMORY;
      return NULL;
    }
    strncat(value, buf, len);
  }

  return NULL;
}


/*****************************************************************************
 *****************************************************************************/

/*
 * FUNCTION NAME:	storeUploadFile
 * DESCRIPTION:		store upload file
 * ARGUMENTS:		name, filename
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
int
storeUploadFile(name, filename)
	const String	name;
	const String	filename;
{
  register FILE *	fin;
  register FILE *	fout;
  char			buf[BUFSIZ];
  size_t		len;

  fin = openUploadFile(name);
  if (fin == NULL) {
    return -1;
  }

  fout = fopen(filename, "w");
  if (fout == NULL) {
    return -1;
  }

  while (!feof(fin)) {
    len = fread(buf, sizeof(char), sizeof(buf), fin);
    fwrite(buf, sizeof(char), len, fout);
  }
  fclose(fout);

  (void)fseek(fin, 0L, SEEK_SET);

  return 0;
}

/*
 * FUNCTION NAME:	digestUploadFile
 * DESCRIPTION:		digest upload file
 * ARGUMENTS:		name, digest_t
 * RETURN VALUE:	Successful: Digest, otherwise: NULL
 */
#ifndef	MD5_DIGEST_LENGTH
#define	MD5_DIGEST_LENGTH	16
#endif	/* MD5_DIGEST_LENGTH */
String
digestUploadFile(fname, type)
	const String	fname;
	digest_t	type;
{
  register FILE *	fin;

  fin = openUploadFile(fname);
  if (fin == NULL) {
    return NULL;
  }

  switch (type) {
    case DIGEST_MD5: {
	MD5_CTX			md5_ctx;
	unsigned char		md5[MD5_DIGEST_LENGTH];
	static unsigned char	md5_hex[MD5_DIGEST_LENGTH*2+1];
	size_t			len;
	char			buf[BUFSIZ];
	register int		i;

	MD5Init(&md5_ctx);
	while ((len = fread(buf, sizeof(char), sizeof(buf), fin)) > 0) {
	  MD5Update(&md5_ctx, buf, len);
	}
	MD5Final(md5, &md5_ctx);

	memset(md5_hex, 0, sizeof(md5_hex));
	for (i = 0; i < MD5_DIGEST_LENGTH; i++) {
	  snprintf(md5_hex + i * 2, sizeof(md5_hex), "%02x", md5[i]);
	}
	(void)fseek(fin, 0L, SEEK_SET);

	return md5_hex;
      }
      break;
    case DIGEST_SHA1: {
	SHA_CTX			sha_ctx;
	unsigned char		sha[SHA_DIGEST_LENGTH];
	static unsigned char	sha_hex[SHA_DIGEST_LENGTH*2+1];
	size_t			len;
	char			buf[BUFSIZ];
	register int		i;

	SHA1_Init(&sha_ctx);
	while ((len = fread(buf, sizeof(char), sizeof(buf), fin)) > 0) {
	  SHA1_Update(&sha_ctx, buf, len);
	}
	SHA1_Final(sha, &sha_ctx);

	memset(sha_hex, 0, sizeof(sha_hex));
	for (i = 0; i < SHA_DIGEST_LENGTH; i++) {
	  snprintf(sha_hex + i * 2, sizeof(sha_hex), "%02x", sha[i]);
	}
	(void)fseek(fin, 0L, SEEK_SET);

	return sha_hex;
      }
      break;
    default:
      break;
  }

  return NULL;
}

